/*
 * Copyright (c) 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

#pragma ident	"@(#)getpeerinfo_svc.c	6.1	99/11/16 SMI; TSOL 2.x"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h> /* getenv, exit */
#include <signal.h>
#include <sys/types.h>
#include <memory.h>
#include <stropts.h>
#include <netconfig.h>
#include <sys/resource.h> /* rlimit */
#include <syslog.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpc/rpc.h>
#include <tsix/t6attrs.h>
#include <tsol/label.h>
#include <tsol/pattr.h>
#include <tsol/priv.h>

#include "getpeerinfo.h"

extern int set_tsol_svc(void);

#ifdef DEBUG
#define	RPC_SVC_FG
#endif

#define	_RPCSVC_CLOSEDOWN 120

#include <sys/param.h>
static int _rpcpmstart;		/* Started by a port monitor ? */

/* States a server can be in wrt request */

#define	_IDLE 0
#define	_SERVED 1

static int _rpcsvcstate = _IDLE;	/* Set when a request is serviced */
static int _rpcsvccount = 0;		/* Number of requests being serviced */

int _debug = 0;			/* set by -D flag */

void
_msgout(msg)
	char *msg;
{
#ifdef RPC_SVC_FG
	if (_rpcpmstart)
		syslog(LOG_ERR, msg);
	else
		(void) fprintf(stderr, "%s\n", msg);
#else
	syslog(LOG_ERR, msg);
#endif
}

static void
closedown(sig)
	int sig;
{
	if (_debug)
		_msgout("closedown");

	if (_rpcsvcstate == _IDLE && _rpcsvccount == 0) {
		extern fd_set svc_fdset;
		static int size;
		int i, openfd;
		struct t_info tinfo;

		if (!t_getinfo(0, &tinfo) && (tinfo.servtype == T_CLTS))
			exit(0);
		if (size == 0) {
			struct rlimit rl;

			rl.rlim_max = 0;
			getrlimit(RLIMIT_NOFILE, &rl);
			if ((size = rl.rlim_max) == 0) {
				return;
			}
		}
		for (i = 0, openfd = 0; i < size && openfd < 2; i++)
			if (FD_ISSET(i, &svc_fdset))
				openfd++;
		if (openfd <= 1)
			exit(0);
	} else
		_rpcsvcstate = _IDLE;

	if (_debug)
		_msgout("closedown not idle waiting again");

	(void) signal(SIGALRM, (void(*)()) closedown);
	(void) alarm(_RPCSVC_CLOSEDOWN/2);
}

static void
peerinfoprog_1(rqstp, transp)
	struct svc_req *rqstp;
	register SVCXPRT *transp;
{
	union {
		peerinfo_arg getpeerinfo_1_arg;
	} argument;
	char *result;
	bool_t (*xdr_argument)(), (*xdr_result)();
	char *(*local)();

	/* allocate space for outgoing attributes, if needed */
	if (transp->xp_tsol_outgoing_attrsp == NULL) {
		t6attr_t	new_incoming;

		/*
		 * First time though. Set to only receive what's needed
		 * and set with SL received.
		 */
		if ((transp->xp_tsol_outgoing_attrsp =
		    t6alloc_blk(T6M_SL)) == NULL) {
			_msgout("unable to allocate "
			    "space for outgoing sl");
			exit(1);
		}
		if ((new_incoming =
		    t6alloc_blk(T6M_SL | T6M_PRIVILEGES)) == NULL) {
			_msgout("unable to allocate "
			    "space for incoming sl and privs");
			exit(1);
		}
		/*
		 * Copy the incoming attrs to be saved to a new incoming
		 * attr block and repalce the old block with the new.
		 */
		if (t6set_attr(T6_SL, t6get_attr(T6_SL,
		    transp->xp_tsol_incoming_attrsp),
		    new_incoming) == NULL) {
			_msgout("unable to copy initial sl");
		}
		if (t6set_attr(T6_PRIVILEGES, t6get_attr(T6_PRIVILEGES,
		    transp->xp_tsol_incoming_attrsp),
		    new_incoming) == NULL) {
			_msgout("unable to copy initial privs");
		}
		t6free_blk(transp->xp_tsol_incoming_attrsp);
		transp->xp_tsol_incoming_attrsp = new_incoming;
	}

	if (_debug) {
		bslabel_t	*slp;
		struct netbuf	*netbuf;
		char		buf[1024];

		/* get the SL */
		if ((slp = (bslabel_t *)t6get_attr(T6_SL,
		    transp->xp_tsol_incoming_attrsp)) == NULL) {
			_msgout("unable to get incoming sl");
			slp = (bslabel_t *)NULL;
		}

		netbuf = svc_getrpccaller(transp);
		(void) sprintf(buf, "got v1 request from %s at %s SL",
		    inet_ntoa(((struct sockaddr_in *)netbuf)->sin_addr),
		    slp == (bslabel_t *)NULL ? "" : bsltoh(slp));
		_msgout(buf);
	}

	/* Set to send received SL. */

	if (t6set_attr(T6_SL, t6get_attr(T6_SL,
	    transp->xp_tsol_incoming_attrsp),
	    transp->xp_tsol_outgoing_attrsp) == -1) {
		_msgout("unable to set outgoing SL");
	}

	_rpcsvccount++;
	switch (rqstp->rq_proc) {
	case NULLPROC:
		if (_debug)
			_msgout("sending NULL reply");
		(void) svc_sendreply(transp, xdr_void,
			(char *)NULL);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;

	case GETPEERINFO:
		xdr_argument = xdr_peerinfo_arg;
		xdr_result = xdr_peerinfo_res;
		local = (char *(*)()) getpeerinfo_1;
		break;

	default:
		svcerr_noproc(transp);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;
	}
	(void) memset((char *)&argument, 0, sizeof (argument));
	if (!svc_getargs(transp, xdr_argument, (char *)&argument)) {
		svcerr_decode(transp);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;
	}
	result = (*local)(&argument, rqstp);

	if (_debug)
		_msgout("sending reply");

	if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, (char *)&argument)) {
		_msgout("unable to free arguments");
		exit(1);
	}
	_rpcsvccount--;
	_rpcsvcstate = _SERVED;
}

static void
peerinfoprog_2(rqstp, transp)
	struct svc_req *rqstp;
	register SVCXPRT *transp;
{
	union {
		peerinfo_arg2 getpeerinfo_2_arg;
	} argument;
	char *result;
	bool_t (*xdr_argument)(), (*xdr_result)();
	char *(*local)();

	/* allocate space for outgoing attributes, if needed */
	if (transp->xp_tsol_outgoing_attrsp == NULL) {
		t6attr_t	new_incoming;

		/*
		 * First time though. Set to only receive what's needed
		 * and set with SL received.
		 */
		if ((transp->xp_tsol_outgoing_attrsp =
		    t6alloc_blk(T6M_SL)) == NULL) {
			_msgout("unable to allocate "
			    "space for outgoing sl");
			exit(1);
		}
		if ((new_incoming =
		    t6alloc_blk(T6M_SL | T6M_PRIVILEGES)) == NULL) {
			_msgout("unable to allocate "
			    "space for incoming sl and privs");
			exit(1);
		}
		/*
		 * Copy the incoming attrs to be saved to a new incoming
		 * attr block and repalce the old block with the new.
		 */
		if (t6set_attr(T6_SL, t6get_attr(T6_SL,
		    transp->xp_tsol_incoming_attrsp),
		    new_incoming) == NULL) {
			_msgout("unable to copy initial sl");
		}
		if (t6set_attr(T6_PRIVILEGES, t6get_attr(T6_PRIVILEGES,
		    transp->xp_tsol_incoming_attrsp),
		    new_incoming) == NULL) {
			_msgout("unable to copy initial privs");
		}
		t6free_blk(transp->xp_tsol_incoming_attrsp);
		transp->xp_tsol_incoming_attrsp = new_incoming;
	}

	if (_debug) {
		bslabel_t	*slp;
		struct netbuf	*netbuf;
		char		buf[1024];

		/* get the SL */
		if ((slp = (bslabel_t *)t6get_attr(T6_SL,
		    transp->xp_tsol_incoming_attrsp)) == NULL) {
			_msgout("unable to get incoming sl");
			slp = (bslabel_t *)NULL;
		}

		netbuf = svc_getrpccaller(transp);
		(void) sprintf(buf, "got v2 request from %s at %s SL",
		    inet_ntoa(((struct sockaddr_in *)netbuf)->sin_addr),
		    slp == (bslabel_t *)NULL ? "" : bsltoh(slp));
		_msgout(buf);
	}

	/* Set to send received SL. */

	if (t6set_attr(T6_SL, t6get_attr(T6_SL,
	    transp->xp_tsol_incoming_attrsp),
	    transp->xp_tsol_outgoing_attrsp) == -1) {
		_msgout("unable to set outgoing SL");
	}

	_rpcsvccount++;
	switch (rqstp->rq_proc) {
	case NULLPROC:
		if (_debug)
			_msgout("sending NULL reply");
		(void) svc_sendreply(transp, xdr_void,
			(char *)NULL);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;

	case GETPEERINFO:
		xdr_argument = xdr_peerinfo_arg2;
		xdr_result = xdr_peerinfo_res;
		local = (char *(*)()) getpeerinfo_2;
		break;

	default:
		svcerr_noproc(transp);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;
	}
	(void) memset((char *)&argument, 0, sizeof (argument));
	if (!svc_getargs(transp, xdr_argument, (char *)&argument)) {
		svcerr_decode(transp);
		_rpcsvccount--;
		_rpcsvcstate = _SERVED;
		return;
	}
	result = (*local)(&argument, rqstp);

	if (_debug)
		_msgout("sending reply");

	if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, (char *)&argument)) {
		_msgout("unable to free arguments");
		exit(1);
	}
	_rpcsvccount--;
	_rpcsvcstate = _SERVED;
}

main(int argc, char *argv[])
{
	pid_t pid;
	int i;
	char mname[FMNAMESZ + 1];
	pattr_flag_t	p_value;
	priv_set_t	pset;
	char		buffer[1024];
	bclabel_t	plabel;
	int		c;

	(void) sigset(SIGPIPE, SIG_IGN);

	openlog("getpeerinfo", LOG_PID, LOG_DAEMON);

	while ((c = getopt(argc, argv, "D")) != EOF)
		switch (c) {
		case 'D':
			_debug++;
			break;
		case '?':
		default:
			_msgout("invalid argument");
			exit(1);
			/* NOTREACHED */
		}

	if (getpattr(PAF_TRUSTED_PATH, &p_value) != 0) {
		(void) sprintf(buffer,
		    "unable to get trusted path process attribute: %s",
		    strerror(errno));
		_msgout(buffer);
		exit(1);
	}
	if (p_value == 0) {
		_msgout("not called from trusted path");
	}
	if (getcmwplabel(&plabel) != 0) {
		(void) sprintf(buffer,
		    "unable to get process label: %s",
		    strerror(errno));
		_msgout(buffer);
		exit(1);
	}
	if (_debug) {
		(void) sprintf(buffer, "starting at %s SL",
		    bsltoh(bcltosl(&plabel)));
		_msgout(buffer);
	}
	/* get the effective privilege set */
	if (getppriv(PRIV_EFFECTIVE, &pset) == -1) {
		(void) sprintf(buffer,
		    "unable to get effective privilege set: %s",
		    strerror(errno));
		_msgout(buffer);
		exit(1);
	}
	if (PRIV_ISASSERT(&pset, PRIV_PROC_SETCLR) != 0) {
		bclear_t	clearance;

		bclearhigh(&clearance);
		if (setclearance(&clearance) != 0) {
			(void) sprintf(buffer,
			    "can't set high clearance %s",
			    strerror(errno));
			_msgout(buffer);
			exit(1);
		}
	}
	if (PRIV_ISASSERT(&pset, PRIV_PROC_SETSL) != 0) {
		bclabel_t	highsl;

		bclhigh(&highsl);
		if (setcmwplabel(&highsl, SETCL_SL) != 0) {
			(void) sprintf(buffer,
			    "can't set high SL %s",
			    strerror(errno));
			_msgout(buffer);
			exit(1);
		}
	}

	/* assert required privileges */
	if ((PRIV_ISASSERT(&pset, PRIV_PROC_AUDIT_TCB) == 0 ||
	    PRIV_ISASSERT(&pset, PRIV_NET_UPGRADE_SL) == 0 ||
	    PRIV_ISASSERT(&pset, PRIV_NET_DOWNGRADE_SL) == 0)) {
		_msgout("required privileges not asserted");
		exit(1);
	}
	/* construct new effective privilege set */
	if (set_effective_priv(PRIV_SET, 3, PRIV_PROC_AUDIT_TCB,
	    PRIV_NET_UPGRADE_SL, PRIV_NET_DOWNGRADE_SL) != 0) {
		_msgout("cannot set effective privilege set");
		exit(1);
	}
	if (_debug) {
		(void) set_effective_priv(PRIV_ON, 1, PRIV_PROC_DUMPCORE);
	}

	if (!ioctl(0, I_LOOK, mname) &&
		(!strcmp(mname, "sockmod") || !strcmp(mname, "timod"))) {
		char *netid;
		struct netconfig *nconf = NULL;
		SVCXPRT *transp;
		int pmclose;

		_rpcpmstart = 1;

		if ((netid = getenv("NLSPROVIDER")) == NULL) {
		/* started from inetd */
			pmclose = 1;
		} else {
			if ((nconf = getnetconfigent(netid)) == NULL)
				_msgout("cannot get transport info");

			pmclose = (t_getstate(0) != T_DATAXFER);
		}
		if (strcmp(mname, "sockmod") == 0) {
			if (ioctl(0, I_POP, 0) || ioctl(0, I_PUSH, "timod")) {
				_msgout("could not get the right module");
				exit(1);
			}
		}
		if ((transp = svc_tli_create(0, nconf, NULL, 0, 0)) == NULL) {
			_msgout("cannot create server handle");
			exit(1);
		}
		if (nconf)
			freenetconfigent(nconf);
		if (!svc_reg(transp, PEERINFOPROG, PEERINFOVERS,
		    peerinfoprog_1, 0)) {
			_msgout("unable to register "
			    "(PEERINFOPROG, PEERINFOVERS).");
			exit(1);
		}
		if (!svc_reg(transp, PEERINFOPROG, PEERINFOVERS2,
		    peerinfoprog_2, 0)) {
			_msgout("unable to register "
			    "(PEERINFOPROG, PEERINFOVERS2).");
			exit(1);
		}
		if (pmclose) {
			(void) signal(SIGALRM, (void(*)()) closedown);
			(void) alarm(_RPCSVC_CLOSEDOWN/2);
		}
		if (set_tsol_svc() == 1) {
			_msgout("couldn't set incoming attr_blk");
			exit(1);
		}
		svc_run();
		exit(1);
		/* NOTREACHED */
	}	else {
#ifndef RPC_SVC_FG
		int size;
		struct rlimit rl;
		pid = fork();
		if (pid < 0) {
			perror("cannot fork");
			exit(1);
		}
		if (pid)
			exit(0);
		rl.rlim_max = 0;
		getrlimit(RLIMIT_NOFILE, &rl);
		if ((size = rl.rlim_max) == 0)
			exit(1);
		for (i = 0; i < size; i++)
			(void) close(i);
		i = open("/dev/console", 2);
		(void) dup2(i, 1);
		(void) dup2(i, 2);
		(void) setsid();
#endif
	}
	if (!svc_create(peerinfoprog_1, PEERINFOPROG, PEERINFOVERS,
	    "netpath")) {
		_msgout("unable to create (PEERINFOPROG, PEERINFOVERS) "
		    "for netpath.");
		exit(1);
	}
	if (!svc_create(peerinfoprog_2, PEERINFOPROG, PEERINFOVERS2,
	    "netpath")) {
		_msgout("unable to create (PEERINFOPROG, PEERINFOVERS2) "
		    "for netpath.");
		exit(1);
	}
	if (set_tsol_svc()) {
		_msgout("couldn't set incoming attr_blk");
		exit(1);
	}

	svc_run();
	_msgout("svc_run returned");
	exit(1);
	/* NOTREACHED */
}
